@@ -1,21 +1,6 @@
-# -*- coding: utf-8 -*-
 """
 The transformer module is for performing cartographic transformations.
-
-Copyright (c) 2019 pyproj Contributors.
-
-Permission to use, copy, modify, and distribute this software
-and its documentation for any purpose and without fee is hereby
-granted, provided that the above copyright notice appear in all
-copies and that both the copyright notice and this permission
-notice appear in supporting documentation. THE AUTHOR DISCLAIMS
-ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT
-SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
-CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
-LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
-NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
-CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE."""
+"""
 
 __all__ = [
     "transform",
@@ -24,15 +9,25 @@ __all__ = [
     "TransformerGroup",
     "AreaOfInterest",
 ]
-
+import warnings
 from array import array
 from itertools import chain, islice
+from pathlib import Path
+from typing import Any, Iterable, Iterator, List, Optional, Tuple, Union
 
-from pyproj import CRS, Proj
-from pyproj._transformer import AreaOfInterest, _Transformer, _TransformerGroup  # noqa
+from pyproj import CRS
+from pyproj._crs import AreaOfUse, CoordinateOperation
+from pyproj._transformer import (  # noqa
+    AreaOfInterest,
+    _Transformer,
+    _TransformerGroup,
+    proj_version_str,
+)
 from pyproj.compat import cstrencode
+from pyproj.datadir import get_user_data_dir
 from pyproj.enums import TransformDirection, WktVersion
 from pyproj.exceptions import ProjError
+from pyproj.sync import _download_resource_file
 from pyproj.utils import _convertback, _copytobuffer
 
 
@@ -54,12 +49,12 @@ class TransformerGroup(_TransformerGroup):
 
     def __init__(
         self,
-        crs_from,
-        crs_to,
-        skip_equivalent=False,
-        always_xy=False,
-        area_of_interest=None,
-    ):
+        crs_from: Any,
+        crs_to: Any,
+        skip_equivalent: bool = False,
+        always_xy: bool = False,
+        area_of_interest: Optional[AreaOfInterest] = None,
+    ) -> None:
         """Get all possible transformations from a :obj:`pyproj.crs.CRS`
         or input used to create one.
 
@@ -82,16 +77,6 @@ class TransformerGroup(_TransformerGroup):
             The area of interest to help order the transformations based on the
             best operation for the area.
 
-
-        Example:
-
-        >>> from pyproj.transformer import TransformerGroup
-        >>> trans_group = TransformerGroup(4326, 2964)
-        >>> trans_group
-        <TransformerGroup: best_available=True>
-        - transformers: 8
-        - unavailable_operations: 1
-
         """
         super().__init__(
             CRS.from_user_input(crs_from),
@@ -104,7 +89,7 @@ class TransformerGroup(_TransformerGroup):
             self._transformers[iii] = Transformer(transformer)
 
     @property
-    def transformers(self):
+    def transformers(self) -> List["Transformer"]:
         """
         list[:obj:`Transformer`]:
             List of available :obj:`Transformer`
@@ -113,7 +98,7 @@ class TransformerGroup(_TransformerGroup):
         return self._transformers
 
     @property
-    def unavailable_operations(self):
+    def unavailable_operations(self) -> List[CoordinateOperation]:
         """
         list[:obj:`pyproj.crs.CoordinateOperation`]:
             List of :obj:`pyproj.crs.CoordinateOperation` that are not
@@ -122,21 +107,64 @@ class TransformerGroup(_TransformerGroup):
         return self._unavailable_operations
 
     @property
-    def best_available(self):
+    def best_available(self) -> bool:
         """
         bool: If True, the best possible transformer is available.
         """
         return self._best_available
 
-    def __repr__(self):
+    def download_grids(
+        self,
+        directory: Optional[Union[str, Path]] = None,
+        open_license: bool = True,
+        verbose: bool = False,
+    ) -> None:
+        """
+        .. versionadded:: 3.0.0
+
+        Download missing grids that can be downloaded automatically.
+
+        .. warning:: There are cases where the URL to download the grid is missing.
+                     In those cases, you can enable enable
+                     :ref:`debugging-internal-proj` and perform a
+                     transformation. The logs will show the grids PROJ searches for.
+
+        Parameters
+        ----------
+        directory: str or Path, optional
+            The directory to download the grids to.
+            Defaults to :func:`pyproj.datadir.get_user_data_dir`
+        open_license: bool, optional
+            If True, will only download grids with an open license.
+            Defaults to False.
+        verbose: bool, optional
+            If True, will print information about grids downloaded.
+            Default is False.
+        """
+        if directory is None:
+            directory = get_user_data_dir(True)
+        for unavailable_operation in self.unavailable_operations:
+            for grid in unavailable_operation.grids:
+                if (
+                    not grid.available
+                    and grid.url.endswith(grid.short_name)
+                    and grid.direct_download
+                    and (grid.open_license or not open_license)
+                ):
+                    _download_resource_file(
+                        file_url=grid.url,
+                        short_name=grid.short_name,
+                        directory=directory,
+                        verbose=verbose,
+                    )
+                elif not grid.available and verbose:
+                    warnings.warn(f"Skipped: {grid}")
+
+    def __repr__(self) -> str:
         return (
-            "<TransformerGroup: best_available={best_available}>\n"
-            "- transformers: {transformers}\n"
-            "- unavailable_operations: {unavailable_operations}"
-        ).format(
-            best_available=self.best_available,
-            transformers=len(self.transformers),
-            unavailable_operations=len(self.unavailable_operations),
+            f"<TransformerGroup: best_available={self.best_available}>\n"
+            f"- transformers: {len(self.transformers)}\n"
+            f"- unavailable_operations: {len(self.unavailable_operations)}"
         )
 
 
@@ -152,7 +180,7 @@ class Transformer:
 
     """
 
-    def __init__(self, base_transformer=None):
+    def __init__(self, base_transformer: Optional[_Transformer] = None) -> None:
         if not isinstance(base_transformer, _Transformer):
             ProjError.clear()
             raise ProjError(
@@ -162,87 +190,107 @@ class Transformer:
         self._transformer = base_transformer
 
     @property
-    def name(self):
+    def name(self) -> str:
         """
         str: Name of the projection.
         """
         return self._transformer.id
 
     @property
-    def description(self):
+    def description(self) -> str:
         """
         str: Description of the projection.
         """
         return self._transformer.description
 
     @property
-    def definition(self):
+    def definition(self) -> str:
         """
         str: Definition of the projection.
         """
         return self._transformer.definition
 
     @property
-    def has_inverse(self):
+    def has_inverse(self) -> bool:
         """
         bool: True if an inverse mapping exists.
         """
         return self._transformer.has_inverse
 
     @property
-    def accuracy(self):
+    def accuracy(self) -> float:
         """
         float: Expected accuracy of the transformation. -1 if unknown.
         """
         return self._transformer.accuracy
 
     @property
-    def area_of_use(self):
+    def area_of_use(self) -> AreaOfUse:
         """
         .. versionadded:: 2.3.0
 
         Returns
         -------
-        AreaOfUse: The area of use object with associated attributes.
+        AreaOfUse:
+            The area of use object with associated attributes.
         """
         return self._transformer.area_of_use
 
     @property
-    def remarks(self):
+    def remarks(self) -> str:
         """
         .. versionadded:: 2.4.0
 
-        str: Remarks about object.
+        Returns
+        -------
+        str:
+            Remarks about object.
         """
         return self._transformer.remarks
 
     @property
-    def scope(self):
+    def scope(self) -> str:
         """
         .. versionadded:: 2.4.0
 
-        str: Scope of object.
+        Returns
+        -------
+        str:
+            Scope of object.
         """
         return self._transformer.scope
 
     @property
-    def operations(self):
+    def operations(self) -> Optional[Tuple[CoordinateOperation]]:
         """
         .. versionadded:: 2.4.0
 
-        tuple[CoordinateOperation]: The operations in a concatenated operation.
+        Returns
+        -------
+        Tuple[CoordinateOperation]:
+            The operations in a concatenated operation.
         """
         return self._transformer.operations
 
+    @property
+    def is_network_enabled(self) -> bool:
+        """
+        .. versionadded:: 3.0.0
+
+        bool:
+            If the network is enabled.
+        """
+        return self._transformer.is_network_enabled
+
     @staticmethod
     def from_proj(
-        proj_from,
-        proj_to,
-        skip_equivalent=False,
-        always_xy=False,
-        area_of_interest=None,
-    ):
-        """Make a Transformer from a :obj:`pyproj.proj.Proj` or input used to create one.
+        proj_from: Any,
+        proj_to: Any,
+        skip_equivalent: bool = False,
+        always_xy: bool = False,
+        area_of_interest: Optional[AreaOfInterest] = None,
+    ) -> "Transformer":
+        """Make a Transformer from a :obj:`pyproj.Proj` or input used to create one.
 
         .. versionadded:: 2.1.2 skip_equivalent
         .. versionadded:: 2.2.0 always_xy
@@ -250,9 +298,9 @@ class Transformer:
 
         Parameters
         ----------
-        proj_from: :obj:`pyproj.proj.Proj` or input used to create one
+        proj_from: :obj:`pyproj.Proj` or input used to create one
             Projection of input data.
-        proj_to: :obj:`pyproj.proj.Proj` or input used to create one
+        proj_to: :obj:`pyproj.Proj` or input used to create one
             Projection of output data.
         skip_equivalent: bool, optional
             If true, will skip the transformation operation if input and output
@@ -267,9 +315,11 @@ class Transformer:
 
         Returns
         -------
-        :obj:`Transformer`
+        Transformer
 
         """
+        from pyproj import Proj
+
         if not isinstance(proj_from, Proj):
             proj_from = Proj(proj_from)
         if not isinstance(proj_to, Proj):
@@ -285,8 +335,12 @@ class Transformer:
 
     @staticmethod
     def from_crs(
-        crs_from, crs_to, skip_equivalent=False, always_xy=False, area_of_interest=None
-    ):
+        crs_from: Any,
+        crs_to: Any,
+        skip_equivalent: bool = False,
+        always_xy: bool = False,
+        area_of_interest: Optional[AreaOfInterest] = None,
+    ) -> "Transformer":
         """Make a Transformer from a :obj:`pyproj.crs.CRS` or input used to create one.
 
         .. versionadded:: 2.1.2 skip_equivalent
@@ -312,7 +366,7 @@ class Transformer:
 
         Returns
         -------
-        :obj:`Transformer`
+        Transformer
 
         """
         return Transformer(
@@ -326,7 +380,7 @@ class Transformer:
         )
 
     @staticmethod
-    def from_pipeline(proj_pipeline):
+    def from_pipeline(proj_pipeline: str) -> "Transformer":
         """Make a Transformer from a PROJ pipeline string.
 
         https://proj.org/operations/pipeline.html
@@ -345,14 +399,14 @@ class Transformer:
 
     def transform(
         self,
-        xx,
-        yy,
-        zz=None,
-        tt=None,
-        radians=False,
-        errcheck=False,
-        direction=TransformDirection.FORWARD,
-    ):
+        xx: Any,
+        yy: Any,
+        zz: Any = None,
+        tt: Any = None,
+        radians: bool = False,
+        errcheck: bool = False,
+        direction: Union[TransformDirection, str] = TransformDirection.FORWARD,
+    ) -> Any:
         """
         Transform points between two coordinate systems.
 
@@ -372,8 +426,8 @@ class Transformer:
         radians: boolean, optional
             If True, will expect input data to be in radians and will return radians
             if the projection is geographic. Default is False (degrees). Ignored for
-            pipeline transformations.
-        errcheck: boolean, optional (default False)
+            pipeline transformations with pyproj 2, but will work in pyproj 3.
+        errcheck: boolean, optional
             If True an exception is raised if the transformation is invalid.
             By default errcheck=False and an invalid transformation
             returns ``inf`` and no exception is raised.
@@ -387,7 +441,7 @@ class Transformer:
         >>> from pyproj import Transformer
         >>> transformer = Transformer.from_crs("epsg:4326", "epsg:3857")
         >>> x3, y3 = transformer.transform(33, 98)
-        >>> "%.3f  %.3f" % (x3, y3)
+        >>> f"{x3:.3f}  {y3:.3f}"
         '10909310.098  3895303.963'
         >>> pipeline_str = (
         ...     "+proj=pipeline +step +proj=longlat +ellps=WGS84 "
@@ -395,8 +449,8 @@ class Transformer:
         ... )
         >>> pipe_trans = Transformer.from_pipeline(pipeline_str)
         >>> xt, yt = pipe_trans.transform(2.1, 0.001)
-        >>> "%.3f  %.3f" % (xt, yt)
-        '120.321  0.057'
+        >>> f"{xt:.3f}  {yt:.3f}"
+        '2.100  0.001'
         >>> transproj = Transformer.from_crs(
         ...     {"proj":'geocent', "ellps":'WGS84', "datum":'WGS84'},
         ...     "EPSG:4326",
@@ -408,7 +462,7 @@ class Transformer:
         ...     3895878.820,
         ...     radians=True,
         ... )
-        >>> "%.3f %.3f %.3f" % (xpj, ypj, zpj)
+        >>> f"{xpj:.3f} {ypj:.3f} {zpj:.3f}"
         '-2.137 0.661 -20.531'
         >>> transprojr = Transformer.from_crs(
         ...     "EPSG:4326",
@@ -416,11 +470,11 @@ class Transformer:
         ...     always_xy=True,
         ... )
         >>> xpjr, ypjr, zpjr = transprojr.transform(xpj, ypj, zpj, radians=True)
-        >>> "%.3f %.3f %.3f" % (xpjr, ypjr, zpjr)
+        >>> f"{xpjr:.3f} {ypjr:.3f} {zpjr:.3f}"
         '-2704026.010 -4253051.810 3895878.820'
         >>> transformer = Transformer.from_proj("epsg:4326", 4326, skip_equivalent=True)
         >>> xeq, yeq = transformer.transform(33, 98)
-        >>> "%.0f  %.0f" % (xeq, yeq)
+        >>> f"{xeq:.0f}  {yeq:.0f}"
         '33  98'
 
         """
@@ -447,23 +501,27 @@ class Transformer:
         )
         # if inputs were lists, tuples or floats, convert back.
         outx = _convertback(xisfloat, xislist, xistuple, inx)
-        outy = _convertback(yisfloat, yislist, xistuple, iny)
+        outy = _convertback(yisfloat, yislist, yistuple, iny)
         return_data = (outx, outy)
         if inz is not None:
-            return_data += (_convertback(zisfloat, zislist, zistuple, inz),)
+            return_data += (  # type: ignore
+                _convertback(zisfloat, zislist, zistuple, inz),
+            )
         if intime is not None:
-            return_data += (_convertback(tisfloat, tislist, tistuple, intime),)
+            return_data += (  # type: ignore
+                _convertback(tisfloat, tislist, tistuple, intime),
+            )
         return return_data
 
     def itransform(
         self,
-        points,
-        switch=False,
-        time_3rd=False,
-        radians=False,
-        errcheck=False,
-        direction=TransformDirection.FORWARD,
-    ):
+        points: Any,
+        switch: bool = False,
+        time_3rd: bool = False,
+        radians: bool = False,
+        errcheck: bool = False,
+        direction: Union[TransformDirection, str] = TransformDirection.FORWARD,
+    ) -> Iterator[Iterable]:
         """
         Iterator/generator version of the function pyproj.Transformer.transform.
 
@@ -483,7 +541,7 @@ class Transformer:
             If True, will expect input data to be in radians and will return radians
             if the projection is geographic. Default is False (degrees). Ignored for
             pipeline transformations.
-        errcheck: boolean, optional (default False)
+        errcheck: boolean, optional
             If True an exception is raised if the transformation is invalid.
             By default errcheck=False and an invalid transformation
             returns ``inf`` and no exception is raised.
@@ -508,7 +566,7 @@ class Transformer:
         >>> pipe_trans = Transformer.from_pipeline(pipeline_str)
         >>> for pt in pipe_trans.itransform([(2.1, 0.001)]):
         ...     '{:.3f} {:.3f}'.format(*pt)
-        '120.321 0.057'
+        '2.100 0.001'
         >>> transproj = Transformer.from_crs(
         ...     {"proj":'geocent', "ellps":'WGS84', "datum":'WGS84'},
         ...     "EPSG:4326",
@@ -580,7 +638,11 @@ class Transformer:
             for pt in zip(*([iter(buff)] * stride)):
                 yield pt
 
-    def to_wkt(self, version=WktVersion.WKT2_2019, pretty=False):
+    def to_wkt(
+        self,
+        version: Union[WktVersion, str] = WktVersion.WKT2_2019,
+        pretty: bool = False,
+    ):
         """
         Convert the projection to a WKT string.
 
@@ -603,11 +665,12 @@ class Transformer:
 
         Returns
         -------
-        str: The WKT string.
+        str:
+            The WKT string.
         """
         return self._transformer.to_wkt(version=version, pretty=pretty)
 
-    def to_json(self, pretty=False, indentation=2):
+    def to_json(self, pretty: bool = False, indentation: int = 2) -> str:
         """
         Convert the projection to a JSON string.
 
@@ -622,11 +685,12 @@ class Transformer:
 
         Returns
         -------
-        str: The JSON string.
+        str:
+            The JSON string.
         """
         return self._transformer.to_json(pretty=pretty, indentation=indentation)
 
-    def to_json_dict(self):
+    def to_json_dict(self) -> dict:
         """
         Convert the projection to a JSON dictionary.
 
@@ -634,41 +698,50 @@ class Transformer:
 
         Returns
         -------
-        dict: The JSON dictionary.
+        dict:
+            The JSON dictionary.
         """
         return self._transformer.to_json_dict()
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.definition
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return (
-            "<{type_name}: {name}>\nDescription: {description}\n"
-            "Area of Use:\n{area_of_use}"
-        ).format(
-            type_name=self._transformer.type_name,
-            name=self.name,
-            description=self.description,
-            area_of_use=self.area_of_use or "- undefined",
+            f"<{self._transformer.type_name}: {self.name}>\n"
+            f"Description: {self.description}\n"
+            f"Area of Use:\n{self.area_of_use or '- undefined'}"
         )
 
+    def __eq__(self, other: Any) -> bool:
+        if not isinstance(other, Transformer):
+            return False
+        return self._transformer.__eq__(other._transformer)
+
+    def is_exact_same(self, other: Any) -> bool:
+        if not isinstance(other, Transformer):
+            return False
+        return self._transformer.is_exact_same(other._transformer)
+
 
 def transform(
-    p1,
-    p2,
-    x,
-    y,
-    z=None,
-    tt=None,
-    radians=False,
-    errcheck=False,
-    skip_equivalent=False,
-    always_xy=False,
+    p1: Any,
+    p2: Any,
+    x: Any,
+    y: Any,
+    z: Any = None,
+    tt: Any = None,
+    radians: bool = False,
+    errcheck: bool = False,
+    skip_equivalent: bool = False,
+    always_xy: bool = False,
 ):
     """
     .. versionadded:: 2.1.2 skip_equivalent
     .. versionadded:: 2.2.0 always_xy
 
+    .. warning:: This function is deprecated. See: :ref:`upgrade_transformer`
+
     x2, y2, z2 = transform(p1, p2, x1, y1, z1)
 
     Transform points between two coordinate systems defined by the
@@ -738,26 +811,37 @@ def transform(
     >>> '%8.3f %8.3f %8.3f %5.3f %5.3f %5.3f' % xy
     ' -92.220  -94.720  -90.370 38.830 39.320 38.750'
     """
+    warnings.warn(
+        (
+            "This function is deprecated. "
+            "See: https://pyproj4.github.io/pyproj/stable/"
+            "gotchas.html#upgrading-to-pyproj-2-from-pyproj-1"
+        ),
+        DeprecationWarning,
+        stacklevel=2,
+    )
     return Transformer.from_proj(
         p1, p2, skip_equivalent=skip_equivalent, always_xy=always_xy
     ).transform(xx=x, yy=y, zz=z, tt=tt, radians=radians, errcheck=errcheck)
 
 
 def itransform(
-    p1,
-    p2,
-    points,
-    switch=False,
-    time_3rd=False,
-    radians=False,
-    errcheck=False,
-    skip_equivalent=False,
-    always_xy=False,
+    p1: Any,
+    p2: Any,
+    points: Iterable[Iterable],
+    switch: bool = False,
+    time_3rd: bool = False,
+    radians: bool = False,
+    errcheck: bool = False,
+    skip_equivalent: bool = False,
+    always_xy: bool = False,
 ):
     """
     .. versionadded:: 2.1.2 skip_equivalent
     .. versionadded:: 2.2.0 always_xy
 
+    .. warning:: This function is deprecated. See: :ref:`upgrade_transformer`
+
     points2 = itransform(p1, p2, points1)
     Iterator/generator version of the function pyproj.transform.
     Transform points between two coordinate systems defined by the
@@ -811,6 +895,15 @@ def itransform(
     '30 60'
 
     """
+    warnings.warn(
+        (
+            "This function is deprecated. "
+            "See: https://pyproj4.github.io/pyproj/stable/"
+            "gotchas.html#upgrading-to-pyproj-2-from-pyproj-1"
+        ),
+        DeprecationWarning,
+        stacklevel=2,
+    )
     return Transformer.from_proj(
         p1, p2, skip_equivalent=skip_equivalent, always_xy=always_xy
     ).itransform(
